#!/usr/bin/python
# imodpy.py module
#
# Authors: Tor Mohling and David Mastronarde
#
# $Id: imodpy.py,v 14546550ad64 2017/11/22 16:33:27 mast $
#
 
"""A collection of useful functions for use in IMOD scripts

This module provides the following functions:
  runcmd(cmd[, input][,outfile][, inStderr]) - spawn a command with optional input,
                                   return its output or print to stdout or a file
  bkgdProcess(commandArr, [outfile] [, errfile] [, returnOnErr]) - Run a process
                                    in background with optional redirection
  getmrcsize(file)     - run the 'header' command on <file>.
                         returns a triple of x,y,z size integers
  getmrc(file, doAll)  - run the 'header' command on <file>.
                         returns a 'tuple' of x,y,z,mode,px,py,pz,
                         plus ox,oy,oz,min,max,mean if doAll is True
  getmrcpixel(file)    - run the 'header' command on <file>
                         returns just a single pixel size, using extended header
                         value if any
  getMontageSize(stack, [plName]) - runs 'montagesize' on <stack>; adds piece 
                                     list file plName if supplied and it exists; 
                                     returns nx, ny, nz in a tuple
  runGoodframe(nx, ny) - runs goodframe with nx and ny, returns two sizes or (-1, -1)
                            or (-2, -2) for error
  getImageFormat(file) - runs header to determine format of <file>; returns
                            TIFF, HDF, likeMRC, or MRC
  datasetFilename(suffix [, root] [, typext]) - makes filename from root and the suffix 
                                                modified with a file type extension if any
  setRootAndExtension(root, extension) - sets default rootname and file type extension to 
                                         use in datasetFilename
  findRootAxisAndExtensions - Determines data set root name, single/dual axis, stack
                              extension type, and com file extension from tilt com file
  makeBackupFile(file)   - renames file to file~, deleting old file~
  exitFromImodError(pn, errout) - prints the error strings in errout and
                                  prepends 'ERROR: pn - ' to the last one
  parselist(line)      -  convert list entry in <line> into list of integers
  readTextFile(filename[ , descrip] [, returnOnErr] [, maxLines) - reads in text file,
                                    strips line endings, returns a list of strings
  writeTextFile(filename, strings [, returnOnErr]) - writes set of strings to a text file
  optionValue(linelist, option, type, ignorecase = False, numVal = 0, otherSep = None) -
                                           finds option value in list of lines
  convertToInteger(valstr, description) - Convert string to int with error 
                                           message if it fails
  completeAndCheckComFile(comfile) - returns complete com file name and root
  cleanChunkFiles(rootname[, logOnly]) - cleans up log and com files from chunks
  cleanupFiles(files) - Removes a list of files with multiple trials if it fails
  extractProgramEntries(comLines, progName, startKey) - extract entries to a program
                                                        from command file lines
  getCygpath(windows, path) - Returns path, or Windows path if windows is not None
  cygwinPath(path)          - Returns path, or a Cygwin path if running in Cygwin

  addIMODbinIgnoreSIGHUP()  - Adds IMOD_DIR/bin to front of PATH and ignores SIGHUP
  printPID(doPrint)   - Prints a PID with system-dependent prefix if doPrint is true
  getIMODversion()    - Returns the IMOD version as a string, or None for error
  imodIsAbsPath(path) - Tests whether the path is an absolute path (works in Cygwin)
  imodAbsPath(path) - Returns absolute path, converted to windows format if on Windows
  newPsutilAPI(version)  - Returns True if psutil is a version with the new API
  imodNice(niceInc) - Sets niceness of process, even on Windows
  imodTempDir() - returns a temporary directory: IMOD_TEMPDIR, /usr/tmp, or /tmp
  setLibPath() - Set path variables for executing Qt programs
  makeCurrentDirWritable([subdir]) - Tries to make sure current directory or given
                               subdirectory is writeable on Windows
  patchSizeFromEntry(patchin) - Gets 3D patch size from letter or three numbers
  autoPatchNumber(size, lower, upper, ifZ, [densityInd]) - Computes number of patches
                               from size and range, depending on axis and density flags
  parallelBoundarySize([default]) - returns size of parallel writing boundary region
                               based on the built-in defaults or an environment variable
  elapsedTimeComponents(startTime) - returns a tuple with minutes, seconds, and tenth of
                                     second since the startTime
  runProcesschunks(comArray, outfile, topCheckFile, proChunkCheckFile, [lookForPause],
                   [sendPauseForF], [printOutput]):  Runs processchunks with additional
                      arguments in comArray, monitoring output and watching for signals
  checkForProChunksQuit(topCheckFile, proChunkCheckFile, [lookForPause], [sendPauseForF])
                        - Checks a file for quit and pause signals and tells processchunks
                            to quit and, optionally, to pause
  fmtstr(string, *args) - formats a string with replacement fields
  prnstr(string, file = sys.stdout, end = '\n', flush = False) - replaces print function
"""

# other modules needed by imodpy
import sys, os, re, time, glob, signal, stat, shlex
from pip import exitError

pyVersion = 100 * sys.version_info[0] + 10 * sys.version_info[1]

if pyVersion < 300:
   import exceptions

# Variables used like an enum for selecting optionValue type
STRING_VALUE, INT_VALUE, FLOAT_VALUE, BOOL_VALUE = 0, 1, 2, 3

# The global place to stash error strings and last exit status
errStrings = []
errStatus = 0
runRetryLimit = 10
runMaxTimeForRetry = 0.5
raiseKeyInterrupt = False
finishSetAndQuit = False
fileTypeExtension = ''
currentRootname = ''

# Use the subprocess module if available except on cygwin where broken
# pipes (early on) occurred occasionally; require it on win32
# popen2 is deprecated in 2.6 onward, and subprocess seems to work now in
# cygwin, so try using it.
useSubprocess = True
if (sys.platform.find('cygwin') < 0 and pyVersion >= 240) or pyVersion >= 260:
   from subprocess import *
else:
   useSubprocess = False
   if sys.platform.find('win32') >= 0:
      prnstr("ERROR: imodpy - Windows Python must be at least version 2.4")
      sys.exit(1)


# define a new exception class so our caller can more easily catch errors
# raised here
class ImodpyError(Exception):
   def __init__(self, args=[]):
      self.args = args
#   def __str__(self):
#      return repr(self.args)


# Use this call to get the error strings from the exception
def getErrStrings():
   return errStrings


# output_lines = imodpy.runcmd(cmd, input_lines, output_file)
#  output_lines, input_lines are ARRAYs; output_lines is None if output_file
# not None
#  cmd is a STRING
#
def runcmd(cmd, input=None, outfile=None, inStderr = None, ignoreStatus = 0):
   """runcmd(cmd[, input][, outfile][, inStderr])
       - spawn a command with optional input, either send its output to
       outfile or return its output in an array of strings.
       cmd is a string; input is an array; outfile is a file object or
       the string 'stdout' to send output to standard out.
       inStderr specifies the destination of standard error output, but it is
       ignored if outfile is 'stdout'.  It can be either a file object, 'stdout'
       for it to be combined with standard output, or 'pipe' for it to be neither
       returned nor written into a file (it will either be lost or appear on the
       terminal depending on whether the subprocess module is being used).
       If the command fails with a broken pipe within 0.5 second,
       it will retry up to 10 times.  Call setRetryLimit() to modify the allowed
       number of retries and the maximum run time for a failure to be retried."""

   global errStrings, errStatus

   cmd = avoidLocalComFile(cmd)

   # Set up flags for whether to collect output or send to stderr
   errStatus = 0
   collect = 1
   toStdout = 0
   output = None
   verbose = os.getenv('RUNCMD_VERBOSE') == '1'
   if verbose:
      prnstr('+++++++++++++++++++++++++')
      prnstr('   runcmd running command:')
      prnstr(cmd)
      if input:
         prnstr('   With input:')
         for l in input:
            prnstr(l)
      sys.stdout.flush()

   if outfile:
      collect = 0
      if isinstance(outfile, str) and outfile == 'stdout':
         toStdout = 1

   retryCount = 0
   while retryCount <= runRetryLimit:
      tryStart = time.time();
      try:
         if useSubprocess:

            # The subprocess interface: input must be all one string
            if input:
               if isinstance(input, list):
                  joined = ''
                  for l in input:
                     joined += l + '\n'
                  input = joined
            else:
               input = None

            if input and pyVersion >= 300 and retryCount == 0:
               input = input.encode()

            if inStderr and isinstance(inStderr, str):
               if inStderr == 'stdout':
                  inStderr = STDOUT
               elif inStderr == 'pipe':
                  inStderr = PIPE

            # Run it three different ways depending on where output goes
            if collect:
               p = Popen(cmd, shell=True, stdout=PIPE, stdin=PIPE, stderr=inStderr)
               kout, kerr = p.communicate(input)
               if pyVersion >= 300:
                  kout = kout.decode()
               if kout:
                  output = kout.splitlines(True)
            elif toStdout:
               p = Popen(cmd, shell=True, stdin=PIPE)
               p.communicate(input)
            else:
               p = Popen(cmd, shell=True, stdout=outfile, stdin=PIPE, stderr=inStderr)
               p.communicate(input)
            ec = p.returncode
            if ec:
               errStatus = ec
            
         else:

            # The old way, use popen, popen2-4 depending on where the output
            # is going
            if toStdout:
               kin = os.popen(cmd, 'w')
               if input:
                  for l in input:
                     prnstr(l, file=kin)
               ec = kin.close()
            else:
               errToFile =inStderr and isinstance(inStderr, file)
               if errToFile:
                  (kin, kout, kerr) = os.popen3(cmd)
               elif inStderr and isinstance(inStderr, str) and inStderr == 'stdout':
                  (kin, kout) = os.popen4(cmd)
               else:
                  (kin, kout) = os.popen2(cmd)
               if input:
                  for l in input:
                     prnstr(l, file=kin)
               kin.close()
               output = kout.readlines()
               kout.close()
               if errToFile:
                  errlines = kerr.readlines()
                  kerr.close()
               kpid, ec = os.wait()
               if not collect and output:
                  outfile.writelines(output)
               if errToFile:
                  inStderr.writelines(errlines)

            # Error status is high byte of os.wait() return value
            if ec:
               errStatus = ec >> 8

         # If it succeeds, break the retry loop
         break

      except KeyboardInterrupt:
         if raiseKeyInterrupt:
            raise KeyboardInterrupt
         sys.exit(1)
      
      except Exception:

         # If it gets a broken pipe after a short enough time,
         # run another trial.  Sleep a bit after trying twice, it helps break the cycle
         # This problem is observed on MAC OSX with commands that take < 0.02 sec,
         # originally with collected output, later with output to file (so what pipe?)
         if str(sys.exc_info()[1]).find('Broken pipe') >= 0 and \
            retryCount < runRetryLimit and time.time() - tryStart < runMaxTimeForRetry:
            retryCount += 1
            if collect:
               prnstr(fmtstr("Retrying " + cmd + " after {:.3f} sec",
                             time.time() - tryStart))
            elif not toStdout and isinstance(outfile, file):
               outfile.truncate()
            if inStderr and useSubprocess and isinstance(inStderr, file):
               inStderr.truncate()
            if retryCount > 1:
               time.sleep(0.1)
         else:
            errStrings = ["command " + cmd + ": " + str(sys.exc_info()[1]) + "\n"]
            raise ImodpyError(errStrings)
      
   if verbose:
      if output:
         prnstr('    Output:')
         for l in output:
            prnstr(l, end='')
      prnstr('-------------------------', flush = True)
      
   if ec and errStatus != ignoreStatus:
      # look thru the output for 'ERROR' line(s) and put them before this
      errstr = cmd + fmtstr(": exited with status {}\n", errStatus)
      errStrings = []
      if collect and output:
         errStrings = [l for l in output
               if l.find('ERROR:') >= 0]
      errStrings.append(errstr)
      raise ImodpyError

   if collect:
      return output
   return None


# Modify the retry limits if necessary
def setRetryLimit(numRetries, maxTime = None):
   """setRetryLimit(numRetries, [maxTime]) - Set max number of retries for runcmd
   with collected output to numRetries (set to 0 to disable retries), and set the
   maximum time after which it will retry to maxTime.
   """
   global runRetryLimit, runMaxTimeForRetry
   runRetryLimit = max(0, numRetries)
   if maxTime:
      runMaxTimeForRetry = maxTime


# Us this call to control whether ctrl C interrupting runcmd will raise an interrupt
def passOnKeyInterrupt(passOn = True):
   global raiseKeyInterrupt
   raiseKeyInterrupt = passOn
   

# Get exit status of last command that was run
def getLastExitStatus():
   return errStatus


# Run a process in background with optional redirection of all output
def bkgdProcess(commandArr, outfile=None, errfile='stdout', returnOnErr = False):
   """bkgdProcess(commandArr, [outfile] [, errfile]) - Run a process in background
   with optional redirection of standard output to a file, and of standard error
   to a separate file or to standard out by default.
   """
   try:

      # If subprocess is allowed, open the files if any
      if useSubprocess:
         outf = None
         errf = None
         errToFile = False
         if errfile == 'stdout':
            errf = STDOUT
         if outfile:
            action = 'Opening ' + outfile + ' for output'
            outf = open(outfile, 'w')
            errToFile = errfile == 'stdout'
         if errfile and errfile != 'stdout':
            action = 'Opening ' + errfile + ' for output'
            errf = open(errfile, 'w')
            errToFile = True

         # Use detached flag on Windows, although it may not be needed
         # In fact, unless stderr is going to a file it keeps it from running there
         action = 'Starting background process ' + commandArr[0]
         if sys.platform.find('win32') >= 0 and errToFile:
            DETACHED_PROCESS = 0x00000008
            Popen(commandArr, shell=False, stdout=outf, stderr=errf,
                  creationflags=DETACHED_PROCESS)
         else:
            Popen(commandArr, shell=False, stdout=outf, stderr=errf)
         return None

      # Otherwise use system call: wrap all args in quotes and add the redirects
      comstr = commandArr[0]
      action = 'Starting background process ' + commandArr[0]
      for i in range(1, len(commandArr)):
         comstr += ' "' + commandArr[i] + '"'
      if outfile:
         comstr += ' > "' + outfile + '"'
      if errfile == 'stdout':
         comstr += ' 2>&1'
      elif errfile:
         comstr += ' 2> "' + errfile + '"'
      comstr += ' &'

      # The return value is not useful when it fails to run
      os.system(comstr)
      return None
         
   except Exception:
      errString = action + "  - " + str(sys.exc_info()[1])
      if returnOnErr:
         return errString
      exitError(errString)


# Get essential data from the header of an MRC file
def getmrc(file, doAll = False):
   """getmrc(file)     - run the 'header' command on <file>
    
    Returns a tuple with seven elements (x,y,z,mode,px,py,pz) by default
    (x,y,z are size in int, mode is int, px,py,pz are pixelsize in float).
    If doAll is True, returns a tuple with 13 elements:
    (x,y,z,mode,px,py,pz,ox,oy,oz,min,max,mean)  where the added elements are
    origin in x, y, z and min, max and mean, all in floats"""

   global errStrings
   input = ["InputFile " + file]
   if doAll:
      hdrout = runcmd("header -si -mo -pi -ori -min -max -mean -StandardInput", input)
      needed = 7
   else:
      hdrout = runcmd("header -si -mo -pi -StandardInput", input)
      needed = 3

   if len(hdrout) < needed:
      errStrings = ["header " + file + ": too few lines of output\n"]
      raise ImodpyError(errStrings)

   nxyz = hdrout[0].split()
   pxyz = hdrout[2].split()
   if doAll:
      orixyz = hdrout[3].split()
   if len(nxyz) < 3 or len(pxyz) < 3 or (doAll and len(orixyz) < 3):
      if len(nxyz) < 3:
         badLine = '-si option: ' + hdrout[0].strip()
      elif  len(pxyz) < 3:
         badLine = '-pi option: ' + hdrout[2].strip()
      else:
         badLine = '-ori option: ' + hdrout[3].strip()
      errStrings = ["header " + file + ": too few numbers on line for " + badLine + "\n"]
      raise ImodpyError(errStrings)
   ix = int(nxyz[0])
   iy = int(nxyz[1])
   iz = int(nxyz[2])
   mode = int(hdrout[1])
   px = float(pxyz[0])
   py = float(pxyz[1])
   pz = float(pxyz[2])

   if not doAll:
      return (ix,iy,iz,mode,px,py,pz)

   orix = float(orixyz[0])
   oriy = float(orixyz[1])
   oriz = float(orixyz[2])
   minv = float(hdrout[4])
   maxv = float(hdrout[5])
   meanv = float(hdrout[6])
   return (ix,iy,iz,mode,px,py,pz,orix,oriy,oriz,minv,maxv,meanv)


# Get size from an MRC file
def getmrcsize(file):
   """getmrcsize(file)    - run the 'header' command on <file>
   
   Returns a triple of x,y,z size integers."""
   (ix,iy,iz,mode,px,py,pz) = getmrc(file)
   return(ix,iy,iz)

# Get one pixel size from an MRC file, based on FEI/Agard value if present
def getmrcpixel(file):
   """getmrcpixel(file)    - run the 'header' command on <file>
   Returns just a single pixel size, based on value in extended header if appropriate"""

   global errStrings
   input = ["InputFile " + file]
   hdrout = runcmd("header -StandardInput", input)
   pixel = -1.
   for line in hdrout:
      try:
         if 'Pixel spacing' in line:
            dotInd = line.find('.. ') + 2
            if dotInd < 3:
               errStrings = ["header " + file + ": cannot find pixel sizes"]
               raise ImodpyError(errStrings)
            lsplit = line[dotInd:].strip().split()
            if len(lsplit) < 3:
               errStrings = ["header " + file + ": pixel sizes not interpretable"]
               raise ImodpyError(errStrings)
            pixel = float(lsplit[0])

         if 'size in nanometers =' in line:
            ind = line.find('=') + 1
            lsplit = line[ind:].strip().split()
            if len(lsplit):
               pixel = 10. * float(lsplit[0])

      except ValueError:
         errStrings = ["header " + file + ": error converting pixel size to float"]
         raise ImodpyError(errStrings)

   if pixel < 0:
      errStrings = ["header " + file + ": cannot find pixel size"]
      raise ImodpyError(errStrings)

   return pixel


# Get the size of a montage by running montagesize
def getMontageSize(stack, plName = None):
   """getMontageSize(stack, [plName])  - runs 'montagesize' on <stack>; adds
   piece list file plName if supplied and it exists; returns nx, ny, nz in a tuple"""
   global errStrings
   command = 'montagesize "' + stack + '"'
   if plName and os.path.exists(plName):
      command += ' "' + plName + '"'
   sizeLines = runcmd(command)
   try:
      problem = 'No output returned'
      line = sizeLines[-1]
      ind = line.find('NZ:')
      if ind < 0:
         problem = 'Line with NZ: not found in output'
         ind = -5
      lsplit = line[ind + 3:].split()
      problem = 'Uninterpretable output on line with NZ:'
      rawXsize = int(lsplit[0])
      rawYsize = int(lsplit[1])
      zsize = int(lsplit[2])
      return (rawXsize, rawYsize, zsize)
   except Exception:
      errStrings = command + ': ' + problem
      raise ImodpyError


# Runs goodframe with the two sizes, nx and ny, and returns the two FFT-compatible sizes
# or (-1, -1) for an error running goodframe, or (-2, -2) for an error in the output
def runGoodframe(nx, ny):
   try:
      goodout = runcmd(fmtstr('goodframe {} {}', nx, ny))
      gsplit = goodout[len(goodout) - 1].split()
      gfnx = int(gsplit[0])
      gfny = int(gsplit[1])
      return (gfnx, gfny)

   except ImodpyError:
      return (-1, -1)
   except Exception:
      return (-2, -2)


# Gets the image format as MRC, HDF, TIFF, or likeMRC
def getImageFormat(file):
   """getImageFormat(file)   - runs header and returns TIFF, HDF, likeMRC, or MRC"""
   retVals = ['TIFF', 'HDF', 'likeMRC']
   nonMRC = ['a TIFF', 'an HDF', 'a non-MRC']
   input = ["InputFile " + file]
   hdrout = runcmd("header -StandardInput", input)
   for ind in range(len(nonMRC)):
      for line in hdrout[0:9]:
         if 'This is ' + nonMRC[ind] in line:
            return retVals[ind]
   return 'MRC'


# Returns a full filename with the root name and suffix if typext is empty, or with the
# last period in the suffix converted to underscore and the typext attached. Suffix should
# include a period and the descriptive extension
def datasetFilename(suffix, root = None, typext = None):
   if root == None:
      root = currentRootname
   if typext == None:
      typext = fileTypeExtension
   if not typext:
      return root + suffix
   for stEnding in ('', '.gz'):
      if suffix.endswith('.st' + stEnding):
         return root + suffix[0:-(len(stEnding) + 2)] + typext + stEnding
   index = suffix.rfind('.')
   if index >= 0:
      suffix = suffix[:index] + '_' + suffix[index + 1:]
   return root + suffix + '.' + typext


# Sets the default rootname and standard extension to use in datasetFilename
def setRootAndExtension(root, extension):
   global fileTypeExtension, currentRootname
   fileTypeExtension = extension
   currentRootname = root


# Determines com file extension, single/dual axis, rootname and extension style by
# looking for and reading tilt command file
# Returns a tuple with (com file extension; 2 if dual axis, 0 if not, -1 if undetermined;
# rootname excluding a/b for dual axis; standard stack extension
# If any string is undetermined an empty string is returned
def findRootAxisAndExtensions():
   # Find tilt com file under either extension and determine axis state
   root = ''
   stackExt = ''
   tiltFile = ''
   dualNum = 0
   for comExt in ('com', 'pcm'):
      if os.path.exists('tilt.' + comExt ):
         single = True
         tiltFile = 'tilt.' + comExt
         break
      if os.path.exists('tilta.' + comExt) and os.path.exists('tiltb.' + comExt):
         dualNum = 2
         tiltFile = 'tilta.' + comExt
         break

   if not tiltFile:
      return ('', -1, '', '')
      
   # Then read the file options from the tilt file
   tiltLines = readTextFile(tiltFile, returnOnErr = True)
   if isinstance(tiltLines, str):
      return (comExt, dualNum, '', '')
   inputLine = optionValue(tiltLines, 'InputProjections', STRING_VALUE)
   outputLine = optionValue(tiltLines, 'OutputFile', STRING_VALUE)
   if inputLine and outputLine:
      (root, outExt) = os.path.splitext(outputLine)
      (root, inExt) = os.path.splitext(inputLine)

      # Make sure the rootname is sensible for dual axis
      if dualNum:
         if root.endswith('a'):
            root = root[:-1]
         else:
            root = ''

      # If the extensions are descriptive style, extension is st, if they match then
      # it is a file-type extension
      if inExt == '.ali' and outExt == '.rec':
         stackExt = 'st'
      elif len(inExt) > 1 and inExt == outExt:
         stackExt = inExt[1:]
      else:
         stackExt = ''

   return (comExt, dualNum, root, stackExt)

   
# Make a backup file
def makeBackupFile(filename):
   """makeBackupFile(file)   - rename file to file~, deleting old file~"""
   if os.path.exists(filename):
      backname = filename + '~'
      try:
         if os.path.exists(backname):
            os.remove(backname)
         os.rename(filename, backname)
      except Exception:
         prnstr(fmtstr('WARNING: Failed to rename existing file {} to {}',
                       filename, backname))


# Print error strings in appropriate format after ImodpyError
def exitFromImodError(pn):
   """exitFromImodError(pn) - prints the error strings in the global errStrings
      and prepends 'ERROR: pn - ' to the last one"""
   line = None
   for l in errStrings:
      if line:
         prnstr(line, end='')
      line = l
   prnstr("ERROR: " + pn + " - " + line, end='')
   sys.exit(1)


# Funtion to parse list of comma-separated ranges
def parselist (line):
   """parselist(line)   - convert list entry in <line> into list of integers
 
   Returns the list of integers, or None for an error. An example of a list is
   1-5,7,9,11,15-20.  Numbers separated by dashes are replaced by
   all of the numbers in the range.  Numbers need not be in any order,
   and backward ranges (10-5) are handled.  Only comma and space
   are valid separators.  A / at the beginning of the string will
   return an error.  Negative numbers can be entered provided
   that the minus sign immediately precedes the number.  E.g.: -3 - -1
   or -3--1 will give -3,-2,-1; -3, -1,1 or -3,-1,1 will give -3,-1,1. """

   list = []
   dashlast = False
   negnum = False
   gotcomma = False
   gotnum = False
   nchars = len(line);

   if not nchars:
      return list
   if (line[0] == '/'):
      return None
   nlist = 0
   ind = 0
   lastnum = 0

   #   find next digit and look for '-', but error out on non -,space
   while (ind < nchars):
      next = line[ind]
      if (next.isdigit()):

         #   got a digit: save ind, find next non-digit
         gotnum = True
         numst = ind
         while (1):
            ind += 1
            next = ''
            if (ind >= nchars):
               break
            next = line[ind]
            if (not next.isdigit()):
               break

         # Convert the number
         if (negnum):
            numst -= 1
         number = int(line[numst:ind])

         # set up loop to add to list
         loopst = number
         idir = 1
         if (dashlast):
            if (lastnum > number):
               idir = -1
            loopst = lastnum + idir
            
         
         i = loopst
         while(idir * i <= idir * number):
            list.append(i)
            nlist += 1
            i += idir

         lastnum = number
         negnum = False
         dashlast = False
         gotcomma = False
         continue
   
      if (next != ',' and next != ' ' and next != '-'):
         return None
      if (next == ','):
         gotcomma = True
      if (next == '-'):
         if (dashlast or (not gotnum) or gotcomma):
            negnum = True
         else:
            dashlast = True
   
      ind += 1

   return list


# Function to read in a text file and strip line endings
def readTextFile(filename, descrip = None, returnOnErr = False, maxLines = None):
   """readTextFile(filename[ , descrip])  - read in text file, strip endings

   Reads in the text file in <filename> if this is a string, or reads from it
   as an open file object if not, strips the line endings and blanks
   from the end of each line, and returns a list of strings.  If maxLines is 
   passed, then it reads up to that number of lines.  Exits with
   exitError on an error opening or reading the file, and adds the optional
   description in <descrip> to the error message.  Or, if returnOnErr is True,
   it returns the error string itself.
   """
   if not descrip:
      descrip = " "
   try:
      errString = "Opening"
      if isinstance(filename, str):
         textfile = open(filename, 'r')
      else:
         textfile = filename
      errString = "Reading"
      if maxLines:
         lines = []
         for ind in range(maxLines):
            line = textfile.readline()
            if not line:
               break
            lines.append(line)
      else:
         lines = textfile.readlines()

   except IOError:
      errString = fmtstr("{} {} {}: {}", errString, descrip, filename, \
                       str(sys.exc_info()[1]))
      if returnOnErr:
         try:
            textfile.close()
         except:
            pass
         return errString
      exitError(errString)

   textfile.close()
   for i in range(len(lines)):
      lines[i] = lines[i].rstrip(' \t\r\n')
   return lines

# Function to write a text file with lines that have no endings
def writeTextFile(filename, strings, returnOnErr = False):
   """writeTextFile(filename, strings) - write set of strings to a text file
   Opens the file <filename> and writes the set of strings in the list
   <strings>, adding a line ending to each.  Exits with exitError upon error,
   or returns an error string if returnOnErr is True.
   """
   try:
      action = 'Opening'
      comf = open(filename, 'w')
      action = 'Writing to'
      if strings:
         for line in strings:
            prnstr(line, file=comf)
      comf.close()
   except IOError:
      errString = action + " file: " + filename + "  - " + str(sys.exc_info()[1])
      if returnOnErr:
         try:
            comf.close()
         except:
            pass
         return errString
      exitError(errString)


# Function to find an option value from a list of strings
def optionValue(linelist, option, type, ignorecase = False, numVal = 0, otherSep = None):
   """optionValue(linelist, option, type[, nocase] [, numVal] [,otherSep]) - get option
   value in strings
   Given a list of strings in <linelist>, searches for one(s) that contain the
   text in <option> and that are not commented out.  A white space separator between the
   option and the value is assumed unless another separator is supplied in optional
   argument <otherSep>. The search is
   case-insensitive if optional argument <nocase> is supplied and is not False
   or None.  The return value depends on the value of <type>:
   0 = STRING_VALUE: a string with white space stripped and a comment removed from the end
   1 = INT_VALUE: an array of integers, or one integer if numVal = 1
   2 = FLOAT_VALUE: an array of floats, or one float if numVal = 1
   3 = BOOL_VALUE: a boolean value, True or false
   It returns all values found if numVal = 0, otherwise it returns just the indicated
   number of values.  Values may be separated by commas or whitespace (commas will be
   converted to spaces before splitting).
   It returns None if the option is not found or if its value is empty, has fewer than
   numVal values, or contains inappropriate characters; in the latter case it issues a
   WARNING:.
   If the option occurs more than once, the latest value applies unless there is an
   error return from the first occurrence.
   """
   flags = 0
   if ignorecase:
      flags = re.IGNORECASE
   sep = r'\s'
   if otherSep:
      sep = r'\s*' + otherSep
   optre = re.compile(r'^\s*' + option, flags)
   subre = re.compile('.*' + option + r'[^\s]*' + sep + r'([^#]*).*', flags)
   comre = re.compile(r'\s*#\s*' + option, flags)
   retval = None
   for line in linelist:
      if optre.search(line) and not comre.search(line):
         valstr = subre.sub(r'\1', line).strip()
         if type > 2:

            # A boolean can be any of these values, but it can also be a line with no
            # separator, which requires a separate re test
            bval = valstr.lower()
            if bval == '0' or bval == 'f' or bval == 'off' or bval == 'false':
               retval = False
            elif bval == '1' or bval == 't' or bval == 'on' or bval == 'true' or \
                   bval == "" or \
                   (not otherSep and re.sub('.*' + option + r'[^\s]*' + r'([^#]*).*', \
                                               r'\1', line).strip() == ""):
               retval = True
            else:
               prnstr("WARNING: optionValue - Boolean entry found with " + \
                     "improper value (" + bval + ") in: " + line)
               
         elif valstr == "":
            prnstr("WARNING: optionValue - No value for option in: " + line)
         elif type <= 0:
            retval = valstr
         else:
            retval = []
            splits = valstr.replace(',', ' ').split()
            try:
               numConv = len(splits)
               if numVal:
                  if numConv < numVal:
                     return None
                  numConv = numVal
               for ind in range(numConv):
                  val = splits[ind]
                  if type == 1:
                     retval.append(int(val))
                  else:
                     retval.append(float(val))
            except Exception:
               prnstr("WARNING: optionValue - Bad character in numeric " + \
                     "entry in: " + line)
               return None

            if numVal == 1:
               retval = retval[0]
               
   return retval


# Convert a string to an integer with an error message if it fails
def convertToInteger(valstr, description):
   try:
      return int(valstr)
   except ValueError:
      exitError('Converting ' + description + ' (' + valstr + ') to integer')
      

# Given a com file entry with optional extension, get the rootname and its
# complete name and check for existence
def completeAndCheckComFile(comfile):
   """completeAndCheckComFile(comfile) - return complete com file name and root
   Given a com file file name that can be file, file., or file.com, compose
   the complete name and the root name and exit with error if file does not
   exist.
   """
   if not comfile:
      exitError("A command file must be entered")

   if comfile.endswith('.com'):
      rootname = comfile[0 : len(comfile) - 4]
   elif comfile.endswith('.'):
      rootname = comfile.rstrip('.')
   else:
      rootname = comfile

   comfile = rootname + '.com'
   if not os.path.exists(comfile):
      exitError("Command file " + comfile + " does not exist")
   return (comfile, rootname)


# Clean up log files from chunks, and com files too if logOnly False
def cleanChunkFiles(rootname, logOnly = False):
   rmlist = glob.glob(rootname + '-[0-9][0-9][0-9]*.log')
   if os.path.exists(rootname + '-start.log'):
      rmlist.append(rootname + '-start.log')
   if os.path.exists(rootname + '-finish.log'):
      rmlist.append(rootname + '-finish.log')
   if not logOnly:
      rmlist += glob.glob(rootname + '-[0-9][0-9][0-9]*.com')
      if os.path.exists(rootname + '-start.com'):
         rmlist.append(rootname + '-start.com')
      if os.path.exists(rootname + '-finish.com'):
         rmlist.append(rootname + '-finish.com')
   try:
      for filename in rmlist:
         os.remove(filename)
   except Exception:
      pass


# Clean up a list of files with multiple trials if it fails
def cleanupFiles(files):

   # Even wih a wait of 0.01 it only took two trials, so try 0.1 for this
   retryWait = 0.1
   maxTrials = 10
   numToDo = len(files)
   stillToDo = numToDo * [1]
   trial = 0
   while trial < maxTrials and numToDo:
      trial += 1
      for ind in range(len(files)):
         if stillToDo[ind]:
            removed = -1
            try:
               os.remove(files[ind])
               removed = 1
            except OSError:
               removed = str(sys.exc_info()[1]).find('No such')
            if removed >= 0:
               stillToDo[ind] = 0
               numToDo -= 1

      if numToDo:
         time.sleep(retryWait)


# Extract the entries to a program from the lines of a command file, looking for a line
# that runs progName and also contains startKey
def extractProgramEntries(comLines, progName, startKey):
   startline = -1
   endline = len(comLines)

   # Find the line that runs the program, then look for a following command
   for i in range(endline):
      line = comLines[i].strip()
      if line.startswith('$'):
         if startline > 0:
            endline = i
            break
         if progName in line and startKey in line:
            startline = i + 1

   if startline < 0:
      return None
   return comLines[startline:endline]


# Returns the Windows path for path using cygpath -m if windows is not None
def getCygpath(windows, path):
   if windows:
      try:
         cygtemp = runcmd('cygpath -m "' + path + '"', inStderr = 'stdout')
         if cygtemp != None and len(cygtemp) > 0:
            return cygtemp[0].strip()
      except Exception:
         pass
   return path


# Returns a Cygwin path for path using cygpath if running in Cygwin
# Falls back to composing the path itself if cygpath fails
def cygwinPath(path):
   if 'cygwin' in sys.platform:
      try:
         cygtemp = runcmd('cygpath "' + path + '"', inStderr = 'stdout')
         if cygtemp != None and len(cygtemp) > 0:
            return cygtemp[0].strip()
      except Exception:
         path = path.replace('\\', '/')
         if  path[1] == ':':
            path = '/cygdrive/' + path[0].lower() + path[2:]
         mess = ''
         if len(errStrings):
            mess = ': ' + errStrings[0].strip()
         prnstr('WARNING: Error running Cygwin program cygpath' + mess + \
                '; returning ' + path)
   return path


# Adds IMOD_DIR/bin to front of PATH and set up to ignore SIGHUP except in Windows
def addIMODbinIgnoreSIGHUP():
   os.environ['PATH'] = os.path.join(cygwinPath(os.environ['IMOD_DIR']), 'bin') + \
       os.pathsep + os.environ['PATH']
   if 'win32' not in sys.platform:
      try:
         signal.signal(signal.SIGHUP, signal.SIG_IGN)
      except Exception:
         pass
   

# Conditionally prints a PID with appropriate OS-dependent prefix
def printPID(doPrint):
   if not doPrint:
      return
   PID = 'Python PID: ' + str(os.getpid()) + '\n'
   if 'win32' in sys.platform:
      PID = 'Windows ' + PID
   elif 'cygwin' in sys.platform:
      PID = 'Cygwin ' + PID
   sys.stderr.write(PID)
   sys.stderr.flush()


# Returns the IMOD version as a string, or None for error
def getIMODversion(): 
   try:
      infoLines = runcmd('imodinfo')
      lsplit = infoLines[0].split()
      return lsplit[2]
   except Exception:
      return None


# Tests whether the path is an absolute path
def imodIsAbsPath(path):
   if 'cygwin' in sys.platform:
      try:
         pathlines = runcmd('cygpath "' + path + '"', inStderr = 'stdout')
         if len(pathlines):
            path = pathlines[0]
      except Exception:
         pass
   return os.path.isabs(path)


# Return an absolute path, converted to windows format if on Windows
def imodAbsPath(path):

   # Cygwin will not work with a windows path, so convert it
   absp = os.path.abspath(cygwinPath(path))
   if 'win32' in sys.platform or 'cygwin' in sys.platform:
      absp = getCygpath(True, absp)
   return absp
      

# Returns whether psutil uses the new API, given the psutil.__version__ as argument
def newPsutilAPI(version):
   try:
      verSplit = version.split('.')
      return int(verSplit[0]) > 1
   except Exception:
      return False


# Function to increment nice value of current process; for Windows Python it uses psutil
# and sets to below normal priority between 4 and 15, idle priority above 15
def imodNice(niceInc):
   if sys.platform.find('win32') < 0:
      os.nice(niceInc)
      return 0
   if niceInc < 4:
      return 0
   try:
      import psutil
      p = psutil.Process(os.getpid())
      if niceInc <= 15:
         priority = psutil.BELOW_NORMAL_PRIORITY_CLASS
      else:
         priority = psutil.IDLE_PRIORITY_CLASS
      if newPsutilAPI(psutil.__version__):
         p.nice(priority)
      else:
         p.nice = priority         
      return 0
   except ImportError:
      return 1

                  
# Function to return the IMOD temporary directory
def imodTempDir():
   """imodTempDir() - returns a temporary directory: IMOD_TMPDIR, /usr/tmp, or /tmp
   Uses IMOD_TMPDIR as a temporary directory if it is defined, exists, and is writable;
   otherwise uses /usr/tmp if it exists and is writable; otherwise uses /tmp if it
   exists and is writable; otherwise returns None.  Paths are converted to Windows paths
   on Windows.
   """
   windows = sys.platform.find('win32') >= 0 or sys.platform.find('cygwin') >= 0
   imodtemp = os.getenv('IMOD_TMPDIR')
   if imodtemp != None:
      imodtemp = getCygpath(windows, imodtemp)
      if os.path.exists(imodtemp) and os.path.isdir(imodtemp) and \
             os.access(imodtemp, os.W_OK):
         return imodtemp
   imodtemp = getCygpath(windows, '/usr/tmp')
   if os.path.exists(imodtemp) and os.access(imodtemp, os.W_OK):
      return imodtemp
   imodtemp = getCygpath(windows, '/tmp')
   if os.path.exists(imodtemp) and os.access(imodtemp, os.W_OK):
      return imodtemp
   return '.'


# Set the appropriate path variables for executing Qt programs if IMOD_QTLIBDIR is defined
def setLibPath():
   if os.getenv('IMOD_QTLIBDIR') == None:
      return
   macosx = sys.platform.find('darwin') >= 0
   mainvar = 'LD_LIBRARY_PATH'
   if macosx:
      mainvar = 'DYLD_LIBRARY_PATH'
   suffix = ''
   if os.getenv(mainvar) != None:
      suffix = os.pathsep + os.environ[mainvar]
   os.environ[mainvar] = os.environ['IMOD_QTLIBDIR'] + os.pathsep + \
       os.path.join(os.environ['IMOD_DIR'], 'lib') + suffix
   if macosx:
      mainvar = 'DYLD_FRAMEWORK_PATH'
      suffix = ''
      if os.getenv(mainvar) != None:
         suffix = os.pathsep + os.environ[mainvar]
      os.environ[mainvar] = os.environ['IMOD_QTLIBDIR'] + suffix


# Function for runcmd to avoid running a .com file in current directory on Windows
def avoidLocalComFile(command):
   if 'win32' not in sys.platform:
      return command

   def isExecutable(fpath):
      return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

   # There is no problem if this command has a path in front of it or if there is no
   # com file in the current directory
   # This split respects quotes around components and strips them
   comSplit = shlex.split(command)
   comstr = comSplit[0]
   (comPath, comName) = os.path.split(comstr)
   if comPath:
      return command
   if not os.path.isfile(comstr + '.com'):
      return command

   # Problem.  Look on the path to find program.  IMOD is already on front...
   for path in os.environ["PATH"].split(os.pathsep):
      path = path.strip('"')
      fullPath = os.path.join(path, comstr)
      if isExecutable(fullPath) or isExecutable(fullPath + '.exe') or \
          isExecutable(fullPath + '.cmd'):

         # Compose a new command with the full path, and all components quoted
         newcom = '"' + os.path.join(path, comstr) + '"'
         for arg in comSplit[1:]:
            newcom += ' "' + arg + '"'
         return newcom

   return command


# Function to try to make current directory writable on Windows or at least make sure
# that it is
def makeCurrentDirWritable(subdir = '.'):
   # On Windows, make sure the current directory is rxw, first try the cygwin command
   # then use the system command.  Windows python may be useless at this
   # When the directory is set -rwx, its running of chmod gives permission denied, it sees
   # 777 from os.stat and os.chmod has no effect
   # 12/13/13: At least in Cygwin, some Win 7 systems have files with --------- for
   # permissions and this test failed with no consequence; so finish with real test
   if not ('win32' in sys.platform  or 'cygwin' in sys.platform):
      return None
   try:
      runcmd('chmod u+rwx ' + subdir, inStderr = 'stdout')
   except ImodpyError:
      mode = stat.S_IMODE(os.stat('.')[stat.ST_MODE])
      try:
         os.chmod('.', mode | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
      except Exception:
         errStr = writeTextFile(subdir + '/writetest.tmp', ['Test for writability'], True)
         cleanupFiles([subdir + '/writetest.tmp'])
         return errStr


# Two functions for managing patch size and number needed by Setupcombine and
# Autopatchfit
PATCHXY = (64, 80, 100, 120)
PATCHZ = (32, 40, 50, 60)
AUTODELPATCHXY = (80, 40)
AUTODELPATCHZ = (30, 20)

# Function to return a patch size from an entry that can be a letter or a size
# in X, Y, Z
def patchSizeFromEntry(patchin):
   indpatch = 'SMLE'.find(patchin.upper())
   err = 0
   if indpatch >= 0 and len(patchin) == 1:
      patchnx = patchny = PATCHXY[indpatch]
      patchnz = PATCHZ[indpatch]
   else:
      psplit = patchin.split(',')
      err = 1
      if len(psplit) == 3:
         try:
            patchnx = int(psplit[0])
            patchny = int(psplit[1])
            patchnz = int(psplit[2])
            err = 0
         except ValueError:
            err = 1
   if err:
      return (0, 0, 0, err)
   return (patchnx, patchny, patchnz, err)

# Function to compute number of patches from size and limits
def autoPatchNumber(size, lower, upper, ifZ, densityInd = 0):
   delta = AUTODELPATCHXY[densityInd]
   if ifZ:
      delta = min(AUTODELPATCHZ[densityInd], 3 * size // 4)
   return int(round(float(upper - lower - size) / delta + 1.))


# Returns the number of pixels for parallel writing boundary regions, either the default
# value here, one supplied by the caller as the argument, or the value of environment
# variable PARALLEL_BOUNDARY_SIZE if it is greater
def parallelBoundarySize(default = 2048):
   valStr = os.getenv('PARALLEL_BOUNDARY_SIZE')
   if valStr:
      try:
         newVal = int(valStr)
         if newVal > default:
            return newVal
      except ValueError:
         pass
   return default
   

# Get elapsed time and break into minutes, seconds, and nearest tenth of second
def elapsedTimeComponents(startTime):
   used = time.time() - startTime
   minutes = int(used / 60.)
   used -= 60 * minutes
   seconds = int(used)
   frac = int(round((used - seconds) * 10))
   return (minutes, seconds, frac)


# Check a program's top-level check file for finishing current operation or quitting.
# Returns empty string if topCheckFile is None, blank, non-existent, empty, or lacks a
# recognized code on the last line; 'F' if it contains F; or 'Q' if it contains Q; 'P' if
# lookForPause is True and there is a P.  It outputs to proChunkCheckFile: Q if there is
# a Q; P if lookForPause is True and there is P; or P if there is an F and sendPauseForF
# is True.  It outputs a message in the latter two cases so that a message appears
# immediately
def checkForProChunksQuit(topCheckFile, proChunkCheckFile, lookForPause = False,
                          sendPauseForF = False):
   global finishSetAndQuit
   pauseMess = 'RECEIVED SIGNAL TO FINISH CURRENT OPERATIONS AND EXIT'
   if not topCheckFile or not os.path.exists(topCheckFile):
      return 0
   checklines = readTextFile(topCheckFile, None, True)
   if isinstance(checklines, str) or len(checklines) < 1:
      return ''
   if checklines[-1].startswith('Q'):
      writeTextFile(proChunkCheckFile, ['Q'], True)
      return 'Q'
   isP = checklines[-1].startswith('P')
   isF = checklines[-1].startswith('F')
   if isF or (lookForPause and isP):
      if isP or sendPauseForF:
         writeTextFile(proChunkCheckFile, ['P'], True)
      if not finishSetAndQuit:
         prnstr(pauseMess, flush = True)
      finishSetAndQuit = True
      return checklines[-1][0]
   return ''


# Runs processchunks in the background with options given in comArray and output to
# outfile.  Periodically checks topCheckFile for quit and finish signals, calling
# checkForProChunksQuit with the passed or default values of lookForPause and
# sendPauseForF.  Prints processchunksd output if printOutput is True
# Adds 'processchunks', '-P', '-g', '-c', and proChunkCheckFile to the front of the
# comArray.  Returns a 5-tuple:
#    1 for an error or quit, or -1 for a problem starting processchunks, generally with
#        a message
#    -1 for a processchunks ERROR, with message; 1 if finished; 0 if not finished;
#        -2 if processchunks got a pause signal
#    'F' if finish signal received, 'P' if paused signal received and lookForPause is
#        True, 'Q' to quit (which may have a message), '' if nothing
#    number of chunks done
#    a message if any
def runProcesschunks(comArray, outfile, topCheckFile, proChunkCheckFile,
                     lookForPause = False, sendPauseForF = False, printOutput = False):
   sleepTime = 0.2
   startingTimeOut = 60.
   readErrorTimeout = 30.
   checkInterval = 5.
   finished = 0
   topQuit = ''
   numDone = 0
   comArray = ['processchunks', '-P', '-g', '-c', proChunkCheckFile] + comArray
   startTime = time.time()
   message = bkgdProcess(comArray, outfile, 'stdout', True)
   if message:
      return (-1, 0, 0, numDone, message)

   # Monitor the log and wait for completion
   elapsed = 0.
   gotLinesAt = -1.
   opened = False
   readOK = False
   gotPID = False
   finished = 0
   error = 0
   checkTime = 0.
   inChunkError = False
   gotDoneSoFar = False
   where = None
   try:
      while True:
         try:

            # Keep track of whether it opened and whether it read lines without an error
            if not opened:
               out = open(outfile, 'r')
               opened = True
            else:
               where = out.tell()
               lines = out.readlines()
               readOK = True
               if lines:

                  # Keep track of last time lines were gotten without error and look for
                  # various terminations
                  gotLinesAt = elapsed
                  where = None
                  for l in lines:
                     if printOutput and gotDoneSoFar:
                        prnstr(l.strip())
                     if 'DONE SO FAR' in l:
                        gotDoneSoFar = True
                        try:
                           numDone = l.split()[0]
                        except ValueError:
                           pass
                     if l.startswith('CHUNK ERROR:'):
                        inChunkError = True
                     elif l.startswith('END CHUNK ERROR'):
                        inChunkError = False
                     elif not inChunkError and l.startswith('ERROR:'):
                        message = 'processchunks ' + l.strip()
                        finished = -1
                        error = 1
                        break
                     elif l.startswith('Finished reassembling'):
                        finished = 1
                        break
                     elif 'retain' in l and 'existing' in l:
                        finished = -2
                        break
                     elif 'PID:' in l:
                        gotPID = True

         except IOError:
            readOK = False
            pass

         if finished:
            break

         # Check for various timeouts and quit
         if not opened and elapsed > startingTimeOut:
            message = 'Timeout occurred before processchunks output file could be ' +\
                'opened for monitoring'
            return (1, finished, topQuit, numDone, message)

         if readOK and gotLinesAt < 0 and elapsed > startingTimeOut:
            return (1, finished, topQuit, numDone,
                    'Timeout occurred before processchunks started')

         if readOK and not gotPID and elapsed > startingTimeOut:
            message = 'Processchunks apparently failed to run; check ' + \
                        imodAbsPath(outfile) + ' for messages'
            return (1, finished, topQuit, numDone, message)

         # If can't get the output, tell processchunks to quit
         if opened and not readOK and \
                elapsed - gotLinesAt > readErrorTimeout:
            writeTextFile(proChunkCheckFile, ['Q'], True)
            time.sleep(5.)
            message = 'Unable to read processchunks output file without an error'
            return (1, finished, topQuit, numDone, message)

         # Check for Q ourselves in case processchunks is lost
         if elapsed - checkTime > checkInterval:
            checkTime = elapsed
            topQuit = checkForProChunksQuit(topCheckFile, proChunkCheckFile, lookForPause,
                                            sendPauseForF)
            if topQuit == 'Q':
               return (1, finished, topQuit, numDone, message)
               
         
         elapsed += sleepTime
         time.sleep(sleepTime)

         # Seek to the last location before trying again if nothing was gotten.  This is
         # a trick from stackoverflow.com and is needed on Mac
         try:
            if where != None:
               out.seek(where)
         except IOError:
            pass

   except KeyboardInterrupt:
      writeTextFile(proChunkCheckFile, ['Q'], True)
      topQuit = 'Q'
      message = 'Detected keyboard interrupt, told processchunks to quit'
      return (1, finished, topQuit, numDone, message)
   
   if opened:
      try:
         out.close()
      except Exception:
         pass
   return (error, finished, topQuit, numDone, message)


# Function to format a string in new format for earlier versions of python
def fmtstr(stringIn, *args):
   """fmtstr(string, *args) - formats a string with replacement fields
   fmtstr(string, arg list) is equivalent to string.format(arg list) in Python
   2.7/3.1 or higher provided that the replacement fields all contain either
   no numeric index or a simple numeric index.  Fields can also contain a
   format specifier after a : character.  In Python 2.6 or 3.0, missing
   numeric indices are inserted and the format function is run on the string.
   For earlier versions of python, the replacement fields are turned into
   %-style formatting entries, and then the string is formatted with the %
   operator.  Double braces become single ones, and a % is allowed in the
   string without being doubled.
   """
   string = stringIn

   # The numeric indices did not become optional until 3.1/2.7
   if pyVersion >= 310 or (pyVersion < 300 and pyVersion >= 270):
      return string.format(*args)

   # Insert numeric indices if needed and make sure all have them or need them
   numToAdd = 0
   ind = 0
   while True:
      ind = string.find('{', ind) + 1
      if ind == 0 or ind == len(string):
         break
      next = string[ind]
      if next == '{':
         ind += 1
         continue
      if next == '}' or next == ':':
         if numToAdd < 0:
            prnstr('ERROR: imodpy.fmtstr - Bad unnumbered replacement field '+\
                   'in:' + stringIn)
            sys.exit(1)
         string = string[0:ind] + str(numToAdd) + string[ind:]
         numToAdd += 1
      else:
         if numToAdd > 0:
            prnstr('ERROR: imodpy.fmtstr - Bad numbered replacement field '+\
                   'in:' + stringIn)
            sys.exit(1)
         numToAdd = -1

   if pyVersion >= 260:
      return string.format(*args)

   # Now have to convert to old style.  First escape any %
   string = string.replace('%', '%%')
   ind = 0
   arglist = []
   while True:
      ind = string.find('{', ind) + 1
      if ind == 0 or ind == len(string):
         break
      next = string[ind]
      if next == '{':
         ind += 1
         continue

      # find the end of the number
      colon = string.find(':', ind)
      brace = string.find('}', ind)
      if brace < 0:
         prnstr('ERROR: imodpy.fmtstr - Unclosed brace in: ' + stringIn)
         sys.exit(1)
      if colon < 0:
         numend = brace
      else:
         numend = min(colon, brace)

      # convert number and get argument on list
      argnum = int(string[ind:numend])
      if numend == brace and not isinstance(args[argnum], str):
         arglist.append(str(args[argnum]))
      else:
         arglist.append(args[argnum])

      # Collect start and end strings if any
      basestr = ''
      endstr = ''
      if ind > 1:
         basestr = string[0:ind-1]
      if brace < len(string) - 1:
         endstr = string[brace+1:]
      
      # If there is no :, replace brace and number with %s
      # If there is a :, add a % and take the format string
      if numend == brace:
         string = basestr + '%s' + endstr
      else:
         string = basestr + '%' + string[colon+1:brace] + endstr

   return string % tuple(arglist)
      

# Function to replace print, with same format as new print function
def prnstr(string, file = sys.stdout, end = '\n', flush = False):
   """prnstr(string, file = sys.stdout, end = '\n', flush = False) - replaces
   print function
   This function can be called like the new print function to write to a file,
   or to stdout by default, and add a line ending by default.  If end='',
   then it will not write a space after writing the line, so it can be used to
   write strings with line endings as the old print did.  If end=' ' it will
   write a space after the string as the new print function does.  If flush
   is True or the file is open in binary mode, flush() is called after the
   write.
   """
   binary = 'b' in str(file.mode)
   if pyVersion >= 300 and binary:
      string = string.encode()
      end = end.encode()
   if end == '':
      file.write(string)
   else:
      file.write(string + end)
   if binary or flush:
      file.flush()
